/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.core.marker.volume;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraft.world.storage.MapStorage;
import net.minecraft.world.storage.WorldSavedData;

import buildcraft.lib.misc.NBTUtilBC;
import buildcraft.lib.net.MessageManager;

public class WorldSavedDataVolumeBoxes extends WorldSavedData {
    private static final String DATA_NAME = "buildcraft_volume_boxes";
    /**
     * Used to assign {@link WorldSavedDataVolumeBoxes#world} to pass it to {@link VolumeBox},
     * as we can't pass it other way ({@link MapStorage} can call only constructor with one {@link String} argument
     * and then it calls NBT deserialization method,
     * giving us no chance to set the {@link WorldSavedDataVolumeBoxes#world} field).
     */
    private static World currentWorld;
    public final World world = currentWorld;
    public final List<VolumeBox> volumeBoxes = new ArrayList<>();

    public WorldSavedDataVolumeBoxes() {
        super(DATA_NAME);
    }

    @SuppressWarnings("unused")
    public WorldSavedDataVolumeBoxes(String name) {
        super(name);
    }

    public VolumeBox getVolumeBoxAt(BlockPos pos) {
        return volumeBoxes.stream().filter(volumeBox -> volumeBox.box.contains(pos)).findFirst().orElse(null);
    }

    public void addVolumeBox(BlockPos pos) {
        volumeBoxes.add(new VolumeBox(world, pos));
    }

    public VolumeBox getVolumeBoxFromId(UUID id) {
        return volumeBoxes.stream().filter(volumeBox -> volumeBox.id.equals(id)).findFirst().orElse(null);
    }

    public VolumeBox getCurrentEditing(EntityPlayer player) {
        return volumeBoxes.stream().filter(volumeBox -> volumeBox.isEditingBy(player)).findFirst().orElse(null);
    }

    public void tick() {
        AtomicBoolean dirty = new AtomicBoolean(false);
        volumeBoxes.stream().filter(VolumeBox::isEditing).forEach(volumeBox -> {
            EntityPlayer player = volumeBox.getPlayer(world);
            if (player == null) {
                volumeBox.pauseEditing();
                dirty.set(true);
            } else {
                AxisAlignedBB oldAabb = volumeBox.box.getBoundingBox();
                volumeBox.box.reset();
                volumeBox.box.extendToEncompass(volumeBox.getHeld());
                BlockPos lookingAt = new BlockPos(
                    player.func_174791_d()
                        .func_72441_c(0, player.func_70047_e(), 0)
                        .func_178787_e(player.func_70040_Z().func_186678_a(volumeBox.getDist()))
                );
                volumeBox.box.extendToEncompass(lookingAt);
                if (!volumeBox.box.getBoundingBox().equals(oldAabb)) {
                    dirty.set(true);
                }
            }
        });
        for (VolumeBox volumeBox : volumeBoxes) {
            List<Lock> locksToRemove = new ArrayList<>(volumeBox.locks).stream()
                .filter(lock -> !lock.cause.stillWorks(world))
                .collect(Collectors.toList());
            if (!locksToRemove.isEmpty()) {
                volumeBox.locks.removeAll(locksToRemove);
                dirty.set(true);
            }
        }
        if (dirty.get()) {
            func_76185_a();
        }
    }

    @Override
    public void func_76185_a() {
        super.func_76185_a();
        MessageManager.sendToDimension(new MessageVolumeBoxes(volumeBoxes), world.field_73011_w.getDimension());
    }

    @SuppressWarnings("NullableProblems")
    @Override
    public NBTTagCompound func_189551_b(NBTTagCompound nbt) {
        nbt.func_74782_a("volumeBoxes", NBTUtilBC.writeCompoundList(volumeBoxes.stream().map(VolumeBox::writeToNBT)));
        return nbt;
    }

    @SuppressWarnings("NullableProblems")
    @Override
    public void func_76184_a(NBTTagCompound nbt) {
        volumeBoxes.clear();
        NBTUtilBC.readCompoundList(nbt.func_74781_a("volumeBoxes"))
            .map(volumeBoxTag -> new VolumeBox(world, volumeBoxTag))
            .forEach(volumeBoxes::add);
    }

    public static WorldSavedDataVolumeBoxes get(World world) {
        if (world.field_72995_K) {
            throw new IllegalArgumentException("Tried to create a world saved data instance on the client!");
        }
        MapStorage storage = world.getPerWorldStorage();
        currentWorld = world;
        WorldSavedDataVolumeBoxes instance = (WorldSavedDataVolumeBoxes)
            storage.func_75742_a(WorldSavedDataVolumeBoxes.class, DATA_NAME);
        if (instance == null) {
            instance = new WorldSavedDataVolumeBoxes();
            storage.func_75745_a(DATA_NAME, instance);
        }
        currentWorld = null;
        return instance;
    }
}
